// Copyright 2024 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package features

import (
	"context"

	"github.com/google/yamlfmt"
)

func MakeFeatureEOFNewline(linebreakStr string) yamlfmt.Feature {
	return yamlfmt.Feature{
		Name:        "EOF Newline",
		AfterAction: eofNewlineFeature(linebreakStr),
	}
}

func eofNewlineFeature(linebreakStr string) yamlfmt.FeatureFunc {
	return func(_ context.Context, content []byte) (context.Context, []byte, error) {
		// This check works in both linebreak modes.
		if len(content) == 0 || content[len(content)-1] != '\n' {
			linebreakBytes := []byte(linebreakStr)
			content = append(content, linebreakBytes...)
		}
		return nil, content, nil
	}
}
