// Copyright 2025 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package kyaml

import (
	"bytes"

	"sigs.k8s.io/yaml/kyaml"
)

const KYAMLFormatterType string = "kyaml"

type KYAMLFormatter struct{}

func (f *KYAMLFormatter) Type() string {
	return KYAMLFormatterType
}

func (f *KYAMLFormatter) Format(input []byte) ([]byte, error) {
	encoder := &kyaml.Encoder{}
	in := bytes.NewReader(input)
	out := bytes.Buffer{}
	if err := encoder.FromYAML(in, &out); err != nil {
		return input, err
	}
	return out.Bytes(), nil
}

func (f *KYAMLFormatter) ConfigMap() (map[string]any, error) {
	return map[string]any{"type": KYAMLFormatterType}, nil
}
