// Copyright 2024 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package basic

import "fmt"

type BasicFormatterError struct {
	err error
}

func (e BasicFormatterError) Error() string {
	return fmt.Sprintf("basic formatter error: %v", e.err)
}

func (e BasicFormatterError) Unwrap() error {
	return e.err
}

// func wrapBasicFormatterError(err error) error {
// 	return BasicFormatterError{err: err}
// }
